////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "bxdffresnelblend.h"
#include "bxdfmicrofacet.h"

CBxDFFresnelBlend::CBxDFFresnelBlend( CMicrofacetDistribution* pDist) : CBxDFBase( static_cast<BxDFType>( BSDF_REFLECTION | BSDF_GLOSSY ) ), m_pMicroDist( pDist )
{
}

CBxDFFresnelBlend::~CBxDFFresnelBlend(void)
{
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CBxDFFresnelBlend::f( const Ray& Out, const Ray& In, const Vector& Normal, const Material& RenderMat, const Color& LightColor, double Power, float Shade /*= 1.0*/ )

// f( Out, In) evaluates the BxDF for the in & outgoing pair of rays
////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  UNUSED( LightColor );

  double          RetDiffuse, RetSpecular, MatSpecular, MatDiff;
  Vector          InDir  = In.dir();
  Vector          OutDir = Out.dir();
  double          DotInHalf;

  MatDiff     = 1.0 - RenderMat.specular();
  MatSpecular = RenderMat.specular();

  // ashikhmin-shirley diffuse model
  // RetDiffuse = MatDiff * ( CColorComponent(1.0, 1.0, 1.0) - MatSpecular) * (28.f / (23.f * M_PI)) * (1 - powf(1 - .5f * fabsf( Normal.DotProd( InDir )), 5)) * (1 - powf(1 - .5f * fabsf( -Normal.DotProd(OutDir) ), 5));
  RetDiffuse = MatDiff * (28.0 / (23.0 * PI)) * (1.0 - pow(1 - .5 * fabs( (Normal.dot( InDir )) ), 5)) * (1.0 - pow(1.0 - 0.5 * fabs( (-Normal.dot(OutDir)) ), 5));

  Vector HalfVec = (-OutDir + InDir);
  HalfVec.normalize();
  DotInHalf = InDir.dot( HalfVec );

  RetSpecular = CFresnelHelpers::SchlickFresnel( MatSpecular, DotInHalf ) * m_pMicroDist->D( HalfVec, Normal ) / ( 8.0 * PI * fabs( DotInHalf ) * std::max( fabs( ( Normal.dot( InDir ) ) ), fabs( ( -Normal.dot( OutDir ) ) ) ) );

  // add specular highlight
  if( MatSpecular > 0) {

    Vector R = InDir - Normal * (2.0 * (InDir.dot( Normal )) );
    double  Dot = (R.dot( OutDir ));            // specular phong term: ks * cos(N,L)^SpecPower
    if( Dot > SMALL3D ) {
      double Spec = Power * ( pow( Dot, RenderMat.shine() ) * MatSpecular * Shade );
      return RetDiffuse + RetSpecular * Spec;
    }

  }

  return RetDiffuse + RetSpecular * MatSpecular * Shade;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CBxDFFresnelBlend::Sample_f( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, Random& Sampler, BxDFType Type, const Color& LightColor /*= Color( 1.0, 1.0, 1.0 )*/, double Power /*= 1.0*/, float Shade /*= 1.0 */ )

// ATTENTION: By convention the direction of the outgoing ray is opposite to the one which is incoming!
//            (Out is pointing towards the face, wherease in is pointing away as we solve the transport
//             equation practically going reverse from out to in).
//
// Sample_f samples an incoming ray matching to the outgoing ray + its Pdf. Furthermore it evaluates the
// BxDF according to In & Out and returns the color. This is used for reflection, refraction etc. which will
// continue recursively.
////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  UNUSED( Shade ); UNUSED( Power ); UNUSED( Type ); UNUSED( LightColor );

  Vector InDir, OutDir;
  double pRand[2];

//   double u1 = Sampler.real();
//   double u2 = Sampler.real();
  Sampler.realqmcND( 2, pRand );

  OutDir = -Out.dir();
  // Cosine-sample the hemisphere, flipping the direction if necessary
  // set up orthogonal sampling basis

  if (pRand[0] < .5) {
    pRand[0] = 2.0 * pRand[0];

    // Cosine-sample the hemisphere, flipping the direction if necessary
    InDir = CBaseSampler::cosineSampleHemisphere( Normal, pRand[0], pRand[1] );

    // reflect the outgoing ray if Out & In are on opposite sides
    if( OutDir.dot( InDir ) < 0 ) {
      InDir = -InDir;
    }
  }
  else {
    pRand[0] = 2.0 * (pRand[0] - 0.5);
    m_pMicroDist->Sample_f( OutDir, InDir, Normal, pRand[0], pRand[1], Pdf);
  }

  // sample incoming ray
	In = Out;
	In.init();
	In.incDepth();	// to avoid infinite loops
	In.push( RaySegment( Out.hit( -SMALL3D3 ), InDir, Out.prevZone() ) );

  if( !CBaseSampler::sameHemisphere( OutDir, InDir ) )
    return 0.0;

  Pdf = GetPDF( OutDir, InDir, Normal);
  assert( Pdf > 0.0 );

  return f( Out, In, Normal, RenderMat );
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CBxDFFresnelBlend::GetPDF( const Vector& Out, const Vector& In, const Vector& Normal ) const

// ATTENTION: The vectors should be given in the canonical coordinate system!
////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  if( !CBaseSampler::sameHemisphere( Out, In))
    return 0.f;

  return .5f * ( Normal.dot(In) / PI + m_pMicroDist->GetPDF( Out, In, Normal) );
}
// vim:ts=2
