////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "lightmap.h"

#include <string>

using namespace std;

CLightMap::CLightMap() : m_bLoaded(false), m_IntegralStep(10), m_Extension( 1.0 ), m_Intensity( 1.0), m_bAttenuate( false )
{
  // set up the canonical projection system U = X axis, V = Y axis
  SetUVProjection( Vector( 1, 0, 0 ), Vector( 0, 1, 0 ) );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CLightMap::~CLightMap(void)

// [5/26/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CLightMap::SampleLightMap( GeometryEngine* pEngine, int SampleNr, const Ray& Out, Ray& LightRay, float& Transmission )

//
// Return the color sampled in the direction of the light ray
// [5/26/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  UNUSED( SampleNr );

  Color           ResColor;
  double          Attenuation = 1.0;
  Vector          LightDir;

  // If the light map was loaded we take the sample from the light map,
  // otherwise assume a virtual luminaire which is simply not black. Thus, ambient occlusion can
  // be calculated even if no light map was loaded.
  if( m_bLoaded ) {

    // sample color & light ray direction
    ResColor = Sample_f( pEngine, Out, SampleNr, LightDir );

    if( m_bAttenuate )
      Attenuation = 1.0 / ( PI * m_Extension  * m_Extension );
    else
      Attenuation = 1.0;

    // set up light ray from current hit position
		LightRay = Out;
		LightRay.use_clip = Out.use_clip;
		RaySegment NewSegment( Out.hit( -SMALL3D3 ), LightDir, Out.prevZone() );
    NewSegment.tmin = SMALL3D3 * Out.segment().tmin;
    NewSegment.tmax = INFINITE;
		LightRay.push( NewSegment );

		// check if the ray fired in a certain direction is blocked & return the transmission (1 = unblocked, 0 = blocked, (0,1) transparent object)
    Transmission = IsBlocked( pEngine, LightRay );

    return ResColor * m_Intensity * Attenuation;
  }

  return Color( 0.0, 0.0, 0.0);   // return something between black & white, depending on the shadowing
}

////////////////////////////////////////////////////////////////////////

bool CLightMap::LoadMap( const string& FileName )

////////////////////////////////////////////////////////////////////////
{
  m_bLoaded = m_LightMap.OpenDIB( FileName.c_str() );
  //m_LightMap.FlipVertically();
  return m_bLoaded;
}

////////////////////////////////////////////////////////////////////////

Color CLightMap::LookUpColor( const Vector& Dir )

////////////////////////////////////////////////////////////////////////
{
  float u, v;

  // transform from world into local coordinate system (Z = up = vertical, X = U = horizontal)
  Vector LocalDir = m_InvProjection * Dir;
  Vector EndPos = LocalDir * m_Extension;

  // Spherical texture mapping
//   v = acos( EndPos.Z / m_Extension ) / M_PI;
//   if( EndPos.Y >= 0)
//     u = acos( EndPos.Z / (m_Extension * sin(M_PI * (v) ) ) ) / (2.0 * M_PI);
//   else
//     u = (M_PI + acos( EndPos.X / (m_Extension* sin(M_PI * (v) ) ) ) ) / (2.0 * M_PI);


  // Angular mapping proposed by P. Debevec
  // u [-1,1] and v[-1,1]

  double LengthInXY = sqrt( EndPos.x * EndPos.x + EndPos.y * EndPos.y );
  if( LengthInXY > 0.0 ) {
    double r = 0.159154943 * acos( EndPos.z ) / LengthInXY;
    u = 0.5 + EndPos.x * r;
    v = 0.5 - EndPos.y * r;
  }
  else {
    u = m_LightMap.Width() / 2;             // this should rarely (or actually never!) happen. But for a workaround we
    v = m_LightMap.Height() / 2;            // fix the uv coordinates to the center of the image
  }

  //u *= u * (m_LightMap.Width()  - 3);
  //v *= v * (m_LightMap.Height() - 3);
  // remap u & v from [-1,1] to [0, Width/Height]
  u *= (u + 1 ) * (m_LightMap.Width() / 2.0 );
  v *= (v + 1 ) * (m_LightMap.Height() / 2.0 );

  int iu = max( 0, min( static_cast<int>(u), m_LightMap.Width() - 2) );
  int iv = max( 0, min( static_cast<int>(v), m_LightMap.Height() - 2) );

//   float tu = u - iu;
//   float tv = v - iv;

  COLORREF ResColor = m_LightMap.GetPixel( iu, iv);     // no bilinear interpolation

  return Color( GetRValue( ResColor) / 255.0, GetGValue( ResColor)  / 255.0, GetBValue( ResColor) / 255.0 );
}

////////////////////////////////////////////////////////////////////////

  Color CLightMap::LightEmission( const Vector& Dir )

////////////////////////////////////////////////////////////////////////
{
  return LookUpColor( Dir ) * m_Intensity;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

float CLightMap::IsBlocked( GeometryEngine* pEngine, Ray& LightRay )

// check if the ray fired in a certain direction is blocked & return the transmission (1 = unblocked, 0 = blocked, (0,1) transparent object)
//
// [5/27/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	// check intersection of light ray
	int oldBodyCheckId = pEngine->incBodyCheckId();
	if( !pEngine->intersectRay( &LightRay, false ) ) {
		// Check if we hit a black hole or an error
		VZone* z = LightRay.hitZone();
		if( z != nullptr && (z->region()->type() == GRegion::Type::NORMAL ||
													z->region()->type() == GRegion::Type::VOXEL) ) {
			pEngine->bodyCheckId( oldBodyCheckId );
			return 0.0;
		}
	}
	pEngine->bodyCheckId( oldBodyCheckId );

  // we might have an intersection with a body that is not in a valid region when the geometry is mathematically
  // incorrect. in that case we simply assume an opaque body
  if( !LightRay.hitZone() || !LightRay.hitRegion() )
    return 1.0;

  // take transparency into account when calculating shadowing factor
	double Transparency = LightRay.hitRegion()->alpha() != 0 ? LightRay.hitRegion()->alpha() / 255.0 : 0.0;
	return 1.0 - Transparency;  // return transmission
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Color CLightMap::Sample_f( GeometryEngine* pEngine, const Ray& Out, int SampleNr, Vector& LightDir)

// [5/27/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  UNUSED( SampleNr );

	// TODO: replace with QMC
  LightDir = CBaseSampler::uniformSampleHemisphere( Out.normal, pEngine->random.real(), pEngine->random.real() );
	return LookUpColor( LightDir );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CLightMap::SetUVProjection( const Vector& UAxis, const Vector& VAxis )

// define which axis in 3D space will be the horizontal one (U) and which one the vertical (V)
// [5/28/2020 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  // the canonical system is that X = U, and Y = V

	Vector vZ = UAxis.cross( VAxis );
  m_Projection.make( UAxis, VAxis, vZ );
  m_Projection.transpose();
  m_InvProjection = m_Projection;
  m_InvProjection.inverse();
}
