////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#pragma once
#ifndef PHYS_MAT_INTERFACE_H
#define PHYS_MAT_INTERFACE_H

#include "bxdfbase.h"
#include "baseintegrator.h"
#include "graph/color.h"
#include "math/randsampler.h"

//
// Interface class of all the physics models used for different materials. This class can act as a
// container of different BxDF's that make up the attributes of different materials.
//
class CPhysMatInterface
{
public:
	CPhysMatInterface( CBaseIntegrator* pIntegrator, CBxDFBase::BxDFType Type, bool bSamplingRequired ) : m_pIntegrator( pIntegrator ), m_Type( Type ), m_bMCSamplingRequired( bSamplingRequired ) {};
	virtual ~CPhysMatInterface( void ) {};

	bool MatchesType( CBxDFBase::BxDFType Flags ) const { return (m_Type & Flags); };
	bool RequiresMCSampling() const { return m_bMCSamplingRequired; };

	// f( Out, In) evaluates the BxDF for the in & outgoing pair of rays
	// Out = ray coming FROM the camera, In = ray going towards the light
	virtual double f( const Ray& Out, const Ray& In, const Vector& Normal, const Material& RenderMat, const Color& LightColor, double Power, float Shade = 1.0 ) = 0;

	// By default only diffuse sampling is performed. This function needs to be overwritten by the actual physical materials that
	// have reflective and probably translucent components!
	virtual double Sample_f( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, Random& Sampler, CBxDFBase::BxDFType Type, const Color& LightColor = Color( 1.0, 1.0, 1.0 ), double Power = 1.0,  float Shade = 1.0 )
	{
		UNUSED( Type );

		Color  SampleColor;
		Vector Intersection = Out.hit( -SMALL3D3 );
		Vector Dir, TmpVec;

		// sample incoming ray
		// Dir = CBaseSampler::cosineSampleHemisphere( Normal, Sampler.real(), Sampler.real() );
		double pRandVec[2];
		Sampler.realqmcND( 2, pRandVec );		// get Niederreiter 2 quasi-random numbers
		Dir = CBaseSampler::cosineSampleHemisphere( Normal, pRandVec[0], pRandVec[1] );

		In.use_clip = Out.use_clip;
		RaySegment new_segment( Intersection, Dir, Out.prevZone() );
		new_segment.tmin = SMALL3D3;
		new_segment.tmax = INFINITE;
		In.push( new_segment );

		In.incDepth();

		// determine pdf which needs to be normalized to the hemi-sphere
		Pdf = 1.0 / PI;
		assert( Pdf > 0.0 );

		// default light color was set to white
		return f( Out, In, Normal, RenderMat, LightColor, Power, Shade );
	}

protected:
	CBaseIntegrator*					m_pIntegrator;
	const CBxDFBase::BxDFType m_Type;
	bool                      m_bMCSamplingRequired;

};


#endif
// vim:ts=2
