/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Adapted from halotrace written by C. Theis, 2007
 *
 * Author:	Christian.Theis@cern.ch
 *
 * Date: 19-Aug-2019
 *
 *
 * Code to construct a hierarchical bounding volume based GRegion object that can be queried efficiently in which actual region a point is located.
 *
 * This version uses a relatively simple midpoint splitting strategy choosing the axis where the bounding box is largest to minimize overlaps of sub-volumes. This gives reasonable trees but they
 * clearly can be optimized. One could think about calculating Morton codes for each volume in the current bounding volume and order them by the Morton codes. The splitting point could then be determined
 * by looking at the index where the highest bit starts to flip for the first time. This is repeated recursively and we would automatically have a quite well balanced tree. This could also be combined
 * with cost functions based on SAH like Pantaleoni et al. did for Weta's Pantaray. To be investigated.
 *
 */

#include "bvh.h"

#include <vector>
#include <cassert>

GRegion* CBoundingVolHierarchy::m_pHitNode = nullptr;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBoundingVolHierarchy::CBoundingVolHierarchy( GeometryEngine* pEngine, std::vector<GRegion*> RegionList, size_t  NrOfNodes, bool bVerbose /* = false */, int LogLevel /*= 0*/ ) :
                                              m_pLeftLeaf(nullptr), m_pRightLeaf(nullptr), m_NrOfLeaves(0), m_bVerbose( bVerbose ), m_LogLevel( LogLevel ),
                                              m_NodeList( RegionList ), m_pEngine( pEngine )

// [8/20/2019 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  Vector CurBoxMin, CurBoxMax, NextBoxMin, NextBoxMax, TmpBoxMin, TmpBoxMax;
  double Pivot;

  m_bVerbose && std::cerr << "BVH: starting construction for " << NrOfNodes << std::endl;
  if( NrOfNodes < 1 )
    throw std::runtime_error( "BVH: insufficient geometry regions to build BV tree" );

  if( NrOfNodes == 1 ) {
    *this = CBoundingVolHierarchy( m_pEngine, m_NodeList[0], m_NodeList[0] );
  }
  if( NrOfNodes == 2 ) {
    *this = CBoundingVolHierarchy( m_pEngine, m_NodeList[0], m_NodeList[1] );
  }

  if( NrOfNodes > 2 ) {
    // Obtain the centroid of the first bounding box that will be used as a pivot for the qsplit routine
    BBox CurBBox = m_NodeList[0]->bbox();
    for( size_t i = 1; i < NrOfNodes; ++i ) {
      BBox NextBBox = m_NodeList[i]->bbox();
      CurBBox = MergeBoundingBoxes( CurBBox, NextBBox);
    }

    // set up the bounding box of the whole tree
    m_BoundingBox = CurBBox;

    Pivot = CurBBox.centroid( cXAxis);

    size_t MidPoint = QSplit( m_NodeList.data(), NrOfNodes, Pivot, cXAxis );

    // Create a new bounding volume & define the next axis as the new split axis
    m_pLeftLeaf = BuildBranch( m_NodeList.data(), MidPoint );
    m_pRightLeaf = BuildBranch( &m_NodeList.data()[MidPoint], NrOfNodes - MidPoint);
  }

  m_bVerbose && std::cerr << "BVH: Total nr of tree leaves = " << m_NrOfLeaves << std::endl;
  m_bVerbose && std::cerr << "BVH: BBox (" << m_BoundingBox.low() << ") -> (" << m_BoundingBox.high() << ")" << std::endl;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBoundingVolHierarchy::CBoundingVolHierarchy( GeometryEngine* pEngine, GRegion* pNode1, GRegion* pNode2 ) : m_pEngine( pEngine )

// [8/20/2019 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  Vector Box1Min, Box2Min, Box1Max, Box2Max;

  m_pLeftLeaf  = pNode1;
  m_pRightLeaf = pNode2;

  BBox Box1 = pNode1->bbox();
  BBox Box2 = pNode2->bbox();

  m_BoundingBox = MergeBoundingBoxes( Box1, Box2 );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CBoundingVolHierarchy::CBoundingVolHierarchy( GeometryEngine* pEngine, GRegion* pNode1, GRegion* pNode2, const BBox& CurBBox ) : m_pEngine( pEngine )

// [8/20/2019 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  m_pLeftLeaf  = pNode1;
  m_pRightLeaf = pNode2;
  m_BoundingBox = CurBBox;
}

/*
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CBoundingVolHierarchy::Intersect( CRay& Ray )

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  bool bIntersectLeftLeaf = false;
  bool bIntersectRightLeaf = false;
  CRay BoundsRay( Ray ), LeftLeafRay( Ray ), RightLeafRay( Ray );
  SGdouble tLeft = 0.0, tRight = 0.0;
  CHierarchyNode* pLeftHitNode = nullptr;
  CHierarchyNode* pRightHitNode = nullptr;
	CHierarchyNode* pHitPrimitive = nullptr;

  if( !IntersectBounds( BoundsRay ) )
    return false;

  if( m_pLeftLeaf ) {
    bIntersectLeftLeaf  = m_pLeftLeaf->Intersect( LeftLeafRay );
    tLeft  = LeftLeafRay.GetMint();
  }
  if( m_pRightLeaf ) {
    bIntersectRightLeaf = m_pRightLeaf->Intersect( RightLeafRay );
    tRight = RightLeafRay.GetMint();
  }

  // determine the closest hit
  if( bIntersectLeftLeaf  ) {
    Ray.SetMint( tLeft );
		if( (pHitPrimitive = LeftLeafRay.GetHitPrimitive()) ) {
			Ray.SetHitPrimitive( pHitPrimitive );
			Ray.SetHitNormal( LeftLeafRay.GetHitNormal() );
		}
  }

  if( bIntersectRightLeaf && tRight < tLeft ) {
    Ray.SetMint( tRight );
		if( (pHitPrimitive = RightLeafRay.GetHitPrimitive()) ) {
			Ray.SetHitPrimitive( pHitPrimitive );
			Ray.SetHitNormal( RightLeafRay.GetHitNormal() );
		}
  }

  return (bIntersectLeftLeaf || bIntersectRightLeaf );
}
*/

/*
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CBoundingVolHierarchy::IntersectBounds( CRay& Ray )

// A BVH node has no real geometry. Thus, it has to reset the hit primitive
//////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  bool bRetVal = CHierarchyNode::IntersectBounds( Ray );
  Ray.SetHitPrimitive( NULL );

  return bRetVal;
}
*/

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

GRegion* CBoundingVolHierarchy::BuildBranch( GRegion** ppNodeList, size_t NrOfSubBranchNodes, int Axis /* = cXAxis */ )

// [8/20/2019 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  BBox   CurBBox, NextBBox;
  double Pivot;

  m_NrOfLeaves++;

  if( NrOfSubBranchNodes == 1)
    return ppNodeList[0];
  else if( NrOfSubBranchNodes == 2 )
    return new CBoundingVolHierarchy( m_pEngine, ppNodeList[0], ppNodeList[1] );

  // Obtain the centroid of the bounding box that will be used as a pivot for the qsplit routine
  CurBBox = ppNodeList[0]->bbox();
  for( size_t i = 1; i < NrOfSubBranchNodes; ++i ) {
    NextBBox = ppNodeList[i]->bbox();
    CurBBox = MergeBoundingBoxes( CurBBox, NextBBox );
  }

  // FIXME
  // determine the longest extension of the bounding box which will yield the splitting axis (= axis where we have the lowest probability to have overlapping bodies)
  // int Axis  = CurBBox.size().dominant();

  Pivot     = CurBBox.centroid( Axis );

  // let's split according to the respective axis
  size_t  MidPoint = QSplit( ppNodeList, NrOfSubBranchNodes, Pivot, Axis);

  if( m_bVerbose && m_LogLevel >= 1 ) {
    std::cerr << "BVH: Splitting axis = " << Axis << ", pivot = " << Pivot << " splitting " << MidPoint << "/" << NrOfSubBranchNodes - MidPoint << std::endl;
  }

  // assemble a new bounding volume while iterating through all axes
  GRegion* pLeftLeaf  = BuildBranch( ppNodeList, MidPoint, (Axis + 1) % 3 );
  GRegion* pRightLeaf = BuildBranch( &ppNodeList[MidPoint], NrOfSubBranchNodes - MidPoint, (Axis + 1) % 3 );

  return new CBoundingVolHierarchy( m_pEngine, pLeftLeaf, pRightLeaf, CurBBox );

  assert( false);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

size_t  CBoundingVolHierarchy::QSplit( GRegion** ppNodeList, size_t NrOfNodes, double PivotVal, int Axis)

// determine the number of nodes from a subset (given by ppStartNode + NrOfNodes) that have a centroid
// coordinate < PivotVal with respect to a specific axis.
// [8/20/2019 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  Vector   BoxMin, BoxMax;
  double   Centroid = 0.0;
  size_t   RetVal  = 0;

  for( size_t i = 0; i < NrOfNodes; ++i) {

    BBox CurBBox = ppNodeList[i]->bbox();

    Centroid = CurBBox.centroid( Axis );
    if( Centroid < PivotVal ) {
      // iter_swap( ppNodeList.begin() + i, ppNodeList.begin() + RetVal );
      std::swap( ppNodeList[i], ppNodeList[RetVal] );
      RetVal++;
    }
  }

  if( RetVal == 0 || RetVal == NrOfNodes )
    RetVal = NrOfNodes / 2;

  return RetVal;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

BBox CBoundingVolHierarchy::MergeBoundingBoxes( const BBox& Box1, const BBox& Box2 )

// [8/20/2019 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  BBox NewBox( Box1 );
  NewBox.Union( Box2 );

  return NewBox;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

std::pair<GRegion*, VZone*> CBoundingVolHierarchy::Contains( const Vector& Point )

// [8/20/2019 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  GRegion* pLeftHitNode = nullptr;
  GRegion* pRightHitNode = nullptr;
  VZone*   pHitZone = nullptr;

  if( !m_BoundingBox.inside( Point ) )
    return std::make_pair( nullptr, nullptr);

  // branch into further BHVs
  if( m_pLeftLeaf && dynamic_cast<CBoundingVolHierarchy*>(m_pLeftLeaf) ) {
    auto HitResult = static_cast<CBoundingVolHierarchy*>(m_pLeftLeaf)->Contains( Point);
    pLeftHitNode = HitResult.first;
    pHitZone = HitResult.second;
  }
  else if( m_pLeftLeaf && m_pLeftLeaf->bbox().inside( Point ) ) { // check actual GRegion at the end of the tree
    pLeftHitNode = m_pLeftLeaf;
    pHitZone = m_pEngine->getRegion( pLeftHitNode->id() )->inside( m_pEngine, Point, Vector( 1.0, 0.0, 0.0) );
    if( !pHitZone )
      pLeftHitNode = nullptr;
  }

  // branch into further BHVs
  if( m_pRightLeaf && dynamic_cast<CBoundingVolHierarchy*>(m_pRightLeaf) ) {
    auto HitResult = static_cast<CBoundingVolHierarchy*>(m_pRightLeaf)->Contains( Point );
    pRightHitNode = HitResult.first;
		pHitZone = HitResult.second;
  }
  // check actual GRegion at the end of the tree
  else if( m_pRightLeaf && m_pRightLeaf->bbox().inside( Point ) ) {
		pRightHitNode = m_pRightLeaf;
		pHitZone = m_pEngine->getRegion( pRightHitNode->id() )->inside( m_pEngine, Point, Vector( 1.0, 0.0, 0.0) );
  }

  if( pLeftHitNode )
    return std::make_pair( pLeftHitNode, pHitZone );

  return std::make_pair( pRightHitNode, pHitZone );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

VZone* CBoundingVolHierarchy::inside( const Vector& CurPoint )

// [8/20/2019 theis]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  auto HitResult = Contains( CurPoint );

   return HitResult.second;
}
