/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Adapted from halotrace written by C. Theis, 2007
 *
 * Author:	Christian.Theis@cern.ch
 *
 * Date: 19-Aug-2019
 */
#ifndef BOUNDING_VOL_HIERARCHY_H
#define BOUNDING_VOL_HIERARCHY_H

#include <vector>

#include "geo.h"
#include "math/bbox.h"
#include "math/vector.h"
#include "geometry/vzone.h"
#include "geometry/engine.h"
#include "geometry/gregion.h"

int QSplit( GRegion** ppNodeList, int NrOfNodes, double PivotVal, int Axis);
void MergeBoundingBoxes( const Vector& Box1Min, const Vector& Box1Max, const Vector& Box2Min, const Vector& Box2Max, Vector& NewBoxMin, Vector& NewBoxMax );

class CBoundingVolHierarchy : public GRegion
{
public:
  enum  { cXAxis = 0, cYAxis = 1, cZAxis = 2 };

  CBoundingVolHierarchy( GeometryEngine* pEngine, std::vector<GRegion*> RegionList, size_t  NrOfNodes, bool bVerbose = false, int LogLevel = 0 );
  CBoundingVolHierarchy( GeometryEngine* pEngine, GRegion* pNode1, GRegion* pNode2 );
  CBoundingVolHierarchy( GeometryEngine* pEngine, GRegion* pNode1, GRegion* pNode2, const BBox& CurBBox );
  virtual ~CBoundingVolHierarchy( void ) = default;

  // copy ctor
  CBoundingVolHierarchy( const CBoundingVolHierarchy& /*rhs*/ ) = default;
  CBoundingVolHierarchy& operator=( const CBoundingVolHierarchy& /*rhs*/ ) = default;

  // bounding box of the whole tree
  virtual BBox bbox() { return m_BoundingBox; }

  // build a branch starting at ppStartNode using NrOfSubBranchNodes which will all be enclosed by one new bounding volume
  virtual GRegion*  BuildBranch( GRegion** ppStartNode, size_t NrOfSubBranchNodes, int Axis = cXAxis);

  // check if the passed point is inside the BVH and return the respective zone or a nullptr
  virtual VZone*    inside( const Vector& CurPoint );

  // determine the number of nodes from a subset (given by ppStartNode + NrOfNodes) that have a centroid coordinate < PivotVal with respect to a specific axis.
  static  size_t    QSplit( GRegion** ppStartNode, size_t NrOfNodes, double PivotVal, int Axis);

  unsigned int      GetTotalNrOfLeaves() const { return m_NrOfLeaves; };

  size_t            GetNrOfRegions() const { return m_NodeList.size(); };

protected:

  // create a union of two bounding boxes
  BBox              MergeBoundingBoxes( const BBox& Box1, const BBox& Box2 );

  // internal function to recursively check if the passed point is contained
  virtual std::pair<GRegion*, VZone*> Contains( const Vector& CurPoint );

public:
  static GRegion* m_pHitNode;   // node that was intersected by the passed ray

protected:
  GRegion*        m_pLeftLeaf;
  GRegion*        m_pRightLeaf;

  BBox            m_BoundingBox;
  unsigned int    m_NrOfLeaves;

  bool            m_bVerbose;
  unsigned int    m_LogLevel;

  std::vector<GRegion*> m_NodeList;

  GeometryEngine* m_pEngine;
};

#endif
