/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:        26 Feb 2013
 */

#ifndef __GEOMETRY_MATERIAL_H
#define __GEOMETRY_MATERIAL_H

#include <map>
#include <string>

#include "os.h"

#define MAT_CACHE         (0x0001)
#define MAT_NO_ANTIALIAS  (0x0002)
#define MAT_TRANSFORM     (0x0004)

#define MAT_TM_DIFF       (0x0008)
#define MAT_TM_SPEC       (0x0010)
#define MAT_TM_TRANS      (0x0020)
#define MAT_TM_AMB        (0x0040)
#define MAT_TM_MAPPING    (0x0078)        /* all bits for mapping */

/* ============================== Material ============================ */
class Material {
protected:
	int	_id;		// material id
	std::string	_name;	// Material name

	// Physics parameters
	double	_density;

	int	_Z;		// atomic number
	int	_A;		// atomic mass
	double	_weight;	// atomic weight
//	Array	_elements;

	// Optical properties
	dword	_diff;		// diffuse color
	double	_spec;		// specular (reflected) color
	double	_shine;		// specular spot exponent
	double	_ior;		// index of refraction
	double	_fuzz;		// Surface fuzz
	//double  _trans;		// transparency [0 = opaque, 1 = fully transparent]
	int     _bxdfIndex;	// index of the BxDF associated with this material

//	Textmap	tm_diff;
//	Textmap	tm_spec;
//	dword	_amb;		// ambient color
//	Textmap	tm_amb;
//	dword	_cshine;	// spec spot color
//	Textmap	tm_trans;

	dword	_flag;		// is this material valid for shadow caching

//	struct t_texture *tex;	// ptr for color texture
//	struct t_bump    *bump;	// ptr for surface normal texture
//	Matrix  matrix;		// transformation matrix

private:
	bool	_used;		// if this material is being used
static	std::map<const std::string, int> bxdf_map;

public:
	Material(const std::string& aname="");
	~Material() {}

	// Access Data
	void	name(const std::string& aname)	{ _name = aname; }

const	std::string&	name()	const	{ return _name; }
const	char*	cname()		const	{ return _name.c_str(); }

	void	id(const int i)		{ _id = i; }
	int	id()		const	{ return _id; }

	double	density()	const	{ return _density; }
	void	density(double d)	{ _density = d; }

	int	Z()		const	{ return _Z; }
	void	Z(int a)		{ _Z = a; }

	int	A()		const	{ return _A; }
	void	A(int a)		{ _A = a; }

	int	N()		const	{ return _A-_Z; }

	double	weight()	const	{ return _weight; }
	void	weight(double a)	{ _weight = a; }

//	dword	ambient()	const	{ return _amb; }
//	void	ambient(dword a)	{ _amb = a; }

	dword	diffuse()	const	{ return _diff; }
	dword	color()		const	{ return _diff; }
	void	diffuse(dword a)	{ _diff = a; }

	double	specular()	const	{ return _spec; }
	void	specular(double a)	{ _spec = Range(0.0,a,1.0); }

	//double alpha()			const { return _trans; }
	//void  alpha( double a ) { transparent( a ); }
	//double	transparent()		const { return _trans; }
	//void	transparent(double a)	      { _trans = Range( 0.0, a, 1.0); }
	double	shine()		const	{ return _shine; }
	void	shine(double a)		{ _shine = Max(0.0,a); }

	//	dword	shineColor()	const	{ return _cshine; }
//	void	shineColor(dword b)	{ _cshine = b; }
//	void	shine(double a, dword b){ _shine = a; _cshine = b; }

	double	ior()		const	{ return _ior; }
	void	ior(double a)		{ _ior  = a; }

	bool	bxdf(int i)		{ _bxdfIndex = i; return InRange(0,i,18); }
	bool	bxdf(const char* type);
	int	bxdf()		const	{ return _bxdfIndex; }
	std::string bxdfType()	const;

	bool	used()		const	{ return _used; }
	void	used()			{ _used = true; }

	double	fuzz()		const	{ return _fuzz; }
	void	fuzz(double a)		{ _fuzz  = Range(0.0,a,3.0); }

	// Flags
	int	flags()		const	{ return _flag; }
	void	noAntialias()		{ _flag |= MAT_NO_ANTIALIAS; }
	void	antialias()		{ _flag &= ~MAT_NO_ANTIALIAS; }
}; // Material

std::ostream& operator << (std::ostream&, const Material&);

#endif
