/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#include "planebody.h"

#include "mesh.h"
#include "cylbody.h"
#include "spherebody.h"
#include "math/vector.h"

// ================================= GPLABody ==================================
/** set */
void GPLABody::set( const double nx, const double ny, const double nz,
		    const double  x, const double  y, const double  z)
{
	assert(type()==Type::PLA);

	Z.set(nx, ny, nz);
	P.set(x, y, z);
	zlen = Z.normalize();
	if (Eq0(zlen,SMALL)) {
		Z = Vector::Zo;
		zlen = 1.0;
		throw std::string("PLA normal is null");
	}
	findXYZ();
} // set

/** set */
void GPLABody::set(const double pos)
{
	switch (type()) {
		case Type::XYP:
			Z = Vector::Zo;
			X = Vector::Xo;
			Y = Vector::Yo;
			P.set(0.0, 0.0, pos);
			break;

		case Type::XZP:
			Z = Vector::Yo;
			X = Vector::Zo;
			Y = Vector::Xo;
			P.set(0.0, pos, 0.0);
			break;

		case Type::YZP:
			Z = Vector::Xo;
			X = Vector::Yo;
			Y = Vector::Zo;
			P.set(pos, 0.0, 0.0);
			break;

		default:
			assert(false);
	}
} // set

/** set */
void GPLABody::set(double *what)
{
	resetBBox();
	round0(what, nwhats(), SMALL);
	if (type()==Type::PLA)
		set(what[0], what[1], what[2], what[3], what[4], what[5]);
	else
		set(what[0]);
} // set

/* --- get --- */
int GPLABody::get(double *what) const
{
	switch (type()) {
		case Type::PLA:
			what[0] = zlen*Z.x;
			what[1] = zlen*Z.y;
			what[2] = zlen*Z.z;
			what[3] = P.x;
			what[4] = P.y;
			what[5] = P.z;
			return 6;

		case Type::XYP:
			what[0] = P.z;
			return 1;

		case Type::XZP:
			what[0] = P.y;
			return 1;

		case Type::YZP:
			what[0] = P.x;
			return 1;

		default:
			return -1;
	}
} // get

/** position
 * @param x,y,z	new position to set
 */
void GPLABody::position(const Point& r)
{
	GBody::position(r);

	switch (type()) {
		case Type::XYP:
			P.x = P.y = 0.0;
			break;

		case Type::XZP:
			P.x = P.z = 0.0;
			break;

		case Type::YZP:
			P.y = P.z = 0.0;
			break;

		default: ;
	}
	Po = P + zlen*Z;
} // position

/** return sign of the w axis that has to be used for move=20 type
 * see below on move
 */
void GPLABody::signMove20(const Point& r, const Vector& w)
{
	Point tr = _hasMatrix? _invMatrix*r : r;
	tr -= P;
	Vector dw = tr ^ w;
	_sign20 = Sign(dw*Z);
} // signMove20

/** move body item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 * @param w	vertical axis of the viewport
 * WARNING: the w is used for the rotation (20)
 *          and should be either viewport w or -w such as
 *                    ((r-p)^w)*Z > 0
 *          The reason is that the system below gives two solutions
 *          depending on the side that the user start to drag the extremity
 *          and the program has to remember the side
 */
void GPLABody::move(int item, const Point& r, const Vector& w)
{
	resetBBox();
	Point tr = _hasMatrix? _invMatrix*r : r;
	tr -= SP;
	if (item==NodeId_axisR) {
		// Rotate plane using r as a Second node lying on the plane
		// Calculate new normal
		// Vector w = SZ ^ tr;
		// Z = tr ^ w;

		// We have to rotate the plane around w of the viewport
		// keeping the dot(W,N')=const
		// Equations to fulfill
		// 1. N'.r - N'.p = 0 <=> N'*tr = 0
		// 2. W.N' = W.N
		// 3. |N'| = |N|

		Point&  d = tr;
		Vector dw = d ^ w;
		double  c = SZ * w;

		double E = d.x*dw.z;
		if (!Eq0(E,SMALL)) {
			double A = c*d.x*d.y;
			double B = d.y*dw.y + d.z*dw.z;
			double C = c*d.x*d.x;
			double D = d.x*dw.y;

			double AA = B*B + D*D + E*E;
			double BB = 2.0 * (A*B + C*D);
			double CC = A*A + C*C - E*E;

			if (Eq0(AA,SMALL)) return;
			double x1, x2;
			int n = quadratic(BB/AA, CC/AA, &x1, &x2, SMALL);
			if (n==0) return;
			Z.z = x1;
			Z.y =  (C + D*Z.z) / E;
			Z.x = -(A + B*Z.z) / E;

			if (Sign(dw * Z)*_sign20 < 0) {
				Z.z = x2;
				Z.y =  (C + D*Z.z) / E;
				Z.x = -(A + B*Z.z) / E;
			}
		} else {
			// rotate x->y->z->x
			E = d.y*dw.x;
			if (!Eq0(E,SMALL)) {
				double A = c*d.y*d.z;
				double B = d.z*dw.z + d.x*dw.x;
				double C = c*d.y*d.y;
				double D = d.y*dw.z;

				double AA = B*B + D*D + E*E;
				double BB = 2.0 * (A*B + C*D);
				double CC = A*A + C*C - E*E;

				if (Eq0(AA,SMALL)) return;
				double x1, x2;
				int n = quadratic(BB/AA, CC/AA, &x1, &x2, SMALL);
				if (n==0) return;
				Z.x = x1;
				Z.z =  (C + D*Z.x) / E;
				Z.y = -(A + B*Z.x) / E;

				if (Sign(dw * Z)*_sign20 < 0) {
					Z.x = x2;
					Z.z =  (C + D*Z.x) / E;
					Z.y = -(A + B*Z.x) / E;
				}
			} else {
				// rotate x->z, y->x, z->y
				E = d.z*dw.y;
				if (Eq0(E,SMALL)) return;
				double A = c*d.z*d.x;
				double B = d.x*dw.x + d.y*dw.y;
				double C = c*d.z*d.z;
				double D = d.z*dw.x;

				double AA = B*B + D*D + E*E;
				double BB = 2.0 * (A*B + C*D);
				double CC = A*A + C*C - E*E;

				if (Eq0(AA,SMALL)) return;
				double x1, x2;
				int n = quadratic(BB/AA, CC/AA, &x1, &x2, SMALL);
				if (n==0) return;
				Z.y = x1;
				Z.x =  (C + D*Z.y) / E;
				Z.z = -(A + B*Z.y) / E;

				if (Sign(dw * Z)*_sign20 < 0) {
					Z.y = x2;
					Z.x =  (C + D*Z.y) / E;
					Z.z = -(A + B*Z.y) / E;
				}
			}
		}
		assert(Eq(Z*w,c,SMALL));

		Z.normalize();
		try {
			findXYZ();
		} catch (std::string& e) {
			/* do nothing) */
		}
	} else
	if (item==NodeId_Pos)
		P = SP + (tr*SZ)*SZ;
	else
		GBody::move(item, r, w);
	checkType();
} // move

/** rotate body around an axis */
void GPLABody::rotate(const double angle, const Vector& axis)
{
	GBody::rotate(angle,axis);
	checkType();
} // rotate

/** check plane type */
void GPLABody::checkType()
{
	// Check type
	switch (Z.direction(SMALL)) {
		case Vector::Direction::X:
		case Vector::Direction::Xn:
			_type = Type::YZP;
			break;
		case Vector::Direction::Y:
		case Vector::Direction::Yn:
			_type = Type::XZP;
			break;
		case Vector::Direction::Z:
		case Vector::Direction::Zn:
			_type = Type::XYP;
			break;
		default:
			_type = Type::PLA;
	}
} // checkType

/** createQuads */
void GPLABody::createQuads()
{
	GBody::createQuads();
	//      Cx     Cy     Cz      C
	addQuad(Z.x, Z.y, Z.z, -Z*P);
} // createQuads

/* createMesh */
void GPLABody::createMesh()
{
	GBody::createMesh();
	bool first = _mesh->isEmpty();
	if (first) _mesh->allocateVertices(10);

	//       ^ Y
	//       |
	//   6---7---8
	//   |   |   |
	//   3---4---5  ---> X
	//   |   |   |
	//   0---1---2

	_mesh->vertex(0) = P - xlen*X - ylen*Y;
	_mesh->vertex(1) = P          - ylen*Y;
	_mesh->vertex(2) = P + xlen*X - ylen*Y;

	_mesh->vertex(3) = P - xlen*X;
	_mesh->vertex(4) = P;
	_mesh->vertex(5) = P + xlen*X;

	_mesh->vertex(6) = P - xlen*X + ylen*Y;
	_mesh->vertex(7) = P          + ylen*Y;
	_mesh->vertex(8) = P + xlen*X + ylen*Y;

	_mesh->vertex(9) = P + zlen*Z;
	_mesh->calcBbox();

	if (first) {
		_mesh->add(0, 1, 3, true,  false, true );
		_mesh->add(1, 4, 3, true,  true,  false);

		_mesh->add(1, 2, 4, true,  false, true );
		_mesh->add(2, 5, 4, true,  true,  false);

		_mesh->add(3, 4, 6, true,  false, true );
		_mesh->add(4, 7, 6, true,  true,  false);

		_mesh->add(4, 5, 7, true,  false, true );
		_mesh->add(5, 8, 7, true,  true,  false);

		_mesh->add(4, 9, true);	// add edge for normal

		_mesh->process();
	}
} // createMesh

/** @return bounding box of body */
void GPLABody::updateBBox() const
{
	_bbox.infinite();

	Point p = GBody::position();

	switch (vectorZ().direction(SMALL)) {
		case Vector::Direction::Xn:
			_bbox.lowx(p.x);
			break;

		case Vector::Direction::X:
			_bbox.highx(p.x);
			break;

		case Vector::Direction::Yn:
			_bbox.lowy(p.y);
			break;

		case Vector::Direction::Y:
			_bbox.highy(p.y);
			break;

		case Vector::Direction::Zn:
			_bbox.lowz(p.z);
			break;

		case Vector::Direction::Z:
			_bbox.highz(p.z);
			break;

		default:
			/*do nothing - infinite*/;
	}
} // bbox

#ifdef OBBOX
/** @return oriented bounding box of body */
void GPLABody::updateOBB(bool in) const
{
	OBBox& bb = _obbox[(int)in];
	bb.infinite();

	bb.P = GBody::position();
	bb.X = vectorX();
	bb.Y = vectorY();
	bb.Z = vectorZ();
	bb.highPt.z = 0.0;
} // updateOBB
#endif

/**
 * @param body	other body to check location against
 * @return location of this body with respect to another body
 *
 * FIXME TODO add bounding box limits!!!!
 */
GBody::Location GPLABody::_locationWrt(const GBody *body) const
{
	if (body->isPlane()) {
		// Check if normals are co-linear
		Vector thisN = vectorZ();
		Vector bodyN = body->vectorZ();

		double ptot2 = thisN.length2()*bodyN.length2();
		if (ptot2<=0.0) return Location::OVERLAP;	// error
		double dot = (thisN*bodyN)/sqrt(ptot2);

		// if |dot|!=1 they will overlap somewhere
		if (Eq(dot,  1.0, SMALL)) {
			// Normals are parallel
			// Two situations AinB or BinA
			// Check if this is inside body or vice versa
			double pos = Q(0)(body->position());
			if (pos < 0.0)
				return Location::BinA;
			else
				return Location::AinB;
		} else
		if (Eq(dot, -1.0, SMALL)) {
			// Normals are anti-parallel
			// Two situations Overlap or Outside
			double pos = Q(0)(body->position());
			if (pos < 0.0)
				// FIXME could be optimized in the level of Zone::optimize
				// but here we can only return an OVERLAP
				return Location::OVERLAP;
			else
				return Location::OUTSIDE;
		} else
			return Location::OVERLAP;

	} else
	if (body->type() == Type::SPH) {
		// Check distance of SPH center to plane vs radius
		const Quad& q = Q(0);
		double d = q(body->position()) / sqrt(Sq(q.Cx) + Sq(q.Cy) + Sq(q.Cz));
		double r = ((const GSPHBody*)body)->R();
		if (d > r)
			return Location::OUTSIDE;
		else
		if (d < -r)
			return Location::BinA;
		else
			return Location::OVERLAP;
	} else
	if (In(body->type(), Type::BOX, Type::RPP, Type::WED, Type::RAW, Type::ARB)) {
		double n = sqrt(Sq(Q(0).Cx) + Sq(Q(0).Cy) + Sq(Q(0).Cz));
		Point  bP = body->position();
		Vector bX = body->vectorX() * body->vectorXlen();
		Vector bY = body->vectorY() * body->vectorYlen();
		Vector bZ = body->vectorZ() * body->vectorZlen();
		Point V[8] = {  bP,       bP+bX,    bP+bY,
				bP+bZ,    bP+bX+bZ, bP+bY+bZ,
				bP+bX+bY, bP+bX+bY+bZ };
		size_t nV = In(body->type(), Type::WED, Type::RAW)? 6 : 8;
		double d = Q(0)(V[0]) / n;
		if (Eq0(d, SMALL)) return Location::OVERLAP;
		bool ins = d < 0.0;
		for (size_t i=1; i<nV; i++) {
			d = Q(0)(V[i]) / n;
			if (Eq0(d, SMALL)) return Location::OVERLAP;
			if (ins) {
				if (d > 0.0) return Location::OVERLAP;
			} else
				if (d < 0.0) return Location::OVERLAP;
		}
		if (ins)
			return Location::BinA;
		else
			return Location::OUTSIDE;
	} else
	if (body->type()==Type::ARB) {
		// FIXME normally this check could be done for every object
		// assuming a safe sagitta distance
		assert(_mesh!=nullptr);
		double n = sqrt(Sq(Q(0).Cx) + Sq(Q(0).Cy) + Sq(Q(0).Cz));
		double d = Q(0)(body->_mesh->vertex(0)) / n;
		if (Eq0(d, SMALL)) return Location::OVERLAP;
		bool ins = d < 0.0;
		for (size_t i=1; i<body->_mesh->nvertices(); i++) {
			d = Q(0)(body->_mesh->vertex(i)) / n;
			if (Eq0(d, SMALL)) return Location::OVERLAP;
			if (ins) {
				if (d > 0.0) return Location::OVERLAP;
			} else
				if (d < 0.0) return Location::OVERLAP;
		}
		if (ins)
			return Location::BinA;
		else
			return Location::OUTSIDE;
	} else
	if (In(body->type(), Type::XCC, Type::YCC, Type::ZCC)) {
		// Verify if axis of cylinder is perpendicular to plane
		Vector thisN = vectorZ();
		Vector bodyZ = body->vectorZ();
		if (Eq0(thisN * bodyZ, SMALL)) {
			// Check distance of cylinder axis to plane
			const Quad& q = Q(0);
			double d = q(body->position()) / sqrt(Sq(q.Cx) + Sq(q.Cy) + Sq(q.Cz));
			double r = ((const GInfEllCylBody*)body)->R();
			if (d > r)
				return Location::OUTSIDE;
			else
			if (d < -r)
				return Location::BinA;
			else
				return Location::OVERLAP;
		} else
			return Location::OVERLAP;
	//} else
	//if (body->type() == Type::QUA) {
	//	Plane: ax + by + cz + d = 0
	//	Substituting z = - (ax+by+d)/c  (if c!=0) to QUA
	//	we get a Conic equation on QUA(x,y)
	//	Check if Conic is degenerate or not!
	} else {
		BBox bb = body->bbox();
		double n = sqrt(Sq(Q(0).Cx) + Sq(Q(0).Cy) + Sq(Q(0).Cz));

		double d = Q(0)(bb.low()) / n;
		if (Eq0(d, SMALL)) return Location::OVERLAP;
		bool ins = d < 0.0;

		Point p(bb.low().x, bb.low().y, bb.high().z);
		d = Q(0)(p) / n;
		if (Eq0(d, SMALL)) return Location::OVERLAP;
		if (ins) {
			if (d > 0.0) return Location::OVERLAP;
		} else
			if (d < 0.0) return Location::OVERLAP;

		p(bb.low().x, bb.high().y, bb.high().z);
		d = Q(0)(p) / n;
		if (Eq0(d, SMALL)) return Location::OVERLAP;
		if (ins) {
			if (d > 0.0) return Location::OVERLAP;
		} else
			if (d < 0.0) return Location::OVERLAP;

		p(bb.low().x, bb.high().y, bb.low().z);
		d = Q(0)(p) / n;
		if (Eq0(d, SMALL)) return Location::OVERLAP;
		if (ins) {
			if (d > 0.0) return Location::OVERLAP;
		} else
			if (d < 0.0) return Location::OVERLAP;

		p(bb.high().x, bb.low().y, bb.low().z);
		d = Q(0)(p) / n;
		if (Eq0(d, SMALL)) return Location::OVERLAP;
		if (ins) {
			if (d > 0.0) return Location::OVERLAP;
		} else
			if (d < 0.0) return Location::OVERLAP;

		p(bb.high().x, bb.low().y, bb.high().z);
		d = Q(0)(p) / n;
		if (Eq0(d, SMALL)) return Location::OVERLAP;
		if (ins) {
			if (d > 0.0) return Location::OVERLAP;
		} else
			if (d < 0.0) return Location::OVERLAP;

		p(bb.high().x, bb.high().y, bb.high().z);
		d = Q(0)(p) / n;
		if (Eq0(d, SMALL)) return Location::OVERLAP;
		if (ins) {
			if (d > 0.0) return Location::OVERLAP;
		} else
			if (d < 0.0) return Location::OVERLAP;

		p(bb.high().x, bb.high().y, bb.low().z);
		d = Q(0)(p) / n;
		if (Eq0(d, SMALL)) return Location::OVERLAP;
		if (ins) {
			if (d > 0.0) return Location::OVERLAP;
		} else
			if (d < 0.0) return Location::OVERLAP;

		if (ins)
			return Location::BinA;
		else
			return Location::OUTSIDE;
	}
	return GBody::_locationWrt(body);
} // _locationWrt
