/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#include "spherebody.h"

#include "line.h"
#include "mesh.h"
#include "cylbody.h"

#define ROUND2INT(x,a)		Eq0((double)(int)x-x, a)? (double)(int)x : x

// ============================== GSphericalBase ===============================
int GSphericalBase::N_SPHERE_LON = 16;
int GSphericalBase::N_SPHERE_LAT =  8;

/** createMesh - create an ellipsoidal mesh for SPH and ELL */
void GSphericalBase::createMesh()
{
	GBody::createMesh();
	bool first = _mesh->isEmpty();

	if (first) _mesh->allocateVertices(N_SPHERE_LON*(N_SPHERE_LAT-1)+2);

//	vertices(N_SPHERE_LON*(N_SPHERE_LAT-1)+2);
//	edges(N_SPHERE_LON * 2*(N_SPHERE_LAT-1));

	// Create Mesh
	const double stepf = PI2 / (double)N_SPHERE_LON;
	const double step0 = PI  / (double)N_SPHERE_LAT;

	int k = 0;
	_mesh->vertex(k++) = P + zlen*Z;
	_mesh->vertex(k++) = P - zlen*Z;

	double ang0 = step0;
	for (int j=0; j<N_SPHERE_LAT-1; j++, ang0 += step0) {
		double c0, s0;
		bsincos(ang0, &s0, &c0);
		double ang = 0.0;
		for (int i=0; i<N_SPHERE_LON; i++, ang += stepf) {
			double c, s;
			bsincos(ang, &s, &c);
			_mesh->vertex(k++) = P +
					(xlen*c*s0) * X +
					(ylen*s*s0) * Y +
					(zlen*c0)   * Z;
			//_mesh->vertex(k++).set(what[0] + R*c*s0,
			//		what[1] + R*s*s0,
			//		what[2] + R*c0);
		}
	}
	_mesh->calcBbox();

	if (first) {
		// Create faces
		const size_t base = 2;
		const size_t apex = 2 + N_SPHERE_LON*(N_SPHERE_LAT-2);
		for (int i=0; i<N_SPHERE_LON; i++) {
			int in = (i==N_SPHERE_LON-1? 0 : i+1); // next i
			_mesh->add(0, base+i,  base+in);
			_mesh->add(1, apex+in, apex+i);
		}

		size_t start = base;
		for (int j=0; j<N_SPHERE_LAT-2; j++, start += N_SPHERE_LON)
			for (int i=0; i<N_SPHERE_LON; i++) {
				int in = (i==N_SPHERE_LON-1? 0 : i+1); // next i
				_mesh->add(start+i, start+i+N_SPHERE_LON, start+in, true, false, true);
				_mesh->add(start+in, start+i+N_SPHERE_LON, start+in+N_SPHERE_LON, false, true, true);
			}

		_mesh->process();
		assert(_mesh->isClosed());
		assert(_mesh->isOrientable());
#if _DEBUG>2
		cout << endl;
		cout << "SPH Mesh ";
		cout << " isClosed=" << _mesh->isClosed();
		cout << " isOrientable=" << _mesh->isOrientable() << endl;
		cout << "SPH volume=" << _mesh->volume() << endl;
#endif
	}
} // createMesh

// ================================= GSPHBody ==================================
/** set */
void GSPHBody::set(const double x, const double y, const double z, const double r)
{
	P.set(x,y,z);
	R(Max(r,0.0));
	if (r<TOOSMALL) throw std::string("Negative or zero radius");
} // set

/** set */
void GSPHBody::set(double* what)
{
	resetBBox();
	round0(what, nwhats(), SMALL);
	set(what[0], what[1], what[2], what[3]);
} // set

/* --- get --- */
int GSPHBody::get(double* what) const
{
	what[0] = P.x;
	what[1] = P.y;
	what[2] = P.z;
	what[3] = R();
	return 4;
} // get

/** move body item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 * @param w	vertical axis of the viewport
 */
void GSPHBody::move(int item, const Point& r, const Vector& w)
{
	resetBBox();
	if (item==1) {
		Point tr = _hasMatrix? _invMatrix*r : r;
		R((tr-SP).length());
	} else
		GBody::move(item, r, w);
} // move

/** createQuads */
void GSPHBody::createQuads()
{
	GBody::createQuads();
	addQuad(Quad(Quad::Type::SPHERE, P, R()));
} // createQuads

/** @return bounding box of body */
void GSPHBody::updateBBox() const
{
	Point p = GBody::position();
	_bbox.add(p.x-R(), p.y-R(), p.z-R());
	_bbox.add(p.x+R(), p.y+R(), p.z+R());
} // bbox

#ifdef OBBOX
/** @return oriented bounding box of body */
void GSPHBody::updateOBB(bool in) const
{
	OBBox& bb = _obbox[(int)in];

	bb.P = GBody::position();

	if (!in) {
		bb.low()  = Point(-R(), -R(), -R());
		bb.high() = Point( R(),  R(),  R());
	}
	else {
		// Add inner points
		double sc  = R()/Sqrt3;
		bb.low()  = Point(-sc, -sc, -sc);
		bb.high() = Point(+sc, +sc, +sc);
	}
} // updateOBB
#endif

/**
 * @param body	other body to check location against
 * @return location of this body with respect to another body
 */
GBody::Location GSPHBody::_locationWrt(const GBody *body) const
{
	if (body->type() == Type::SPH) {
		// Minimum distance between two centers
		double dist = (position() - body->position()).length();

		if (dist + ((const GSPHBody*)body)->R() <  R())
			return Location::BinA;
		else
		if (dist - ((const GSPHBody*)body)->R() < -R())
			return Location::AinB;
		else
		if (dist - ((const GSPHBody*)body)->R() >  R())
			return Location::OUTSIDE;
		else
			return Location::OVERLAP;
	} else
	if (body->type() == Type::XCC || body->type() == Type::YCC ||
	    body->type() == Type::ZCC) {
		// Minimum distance between center and axes
		double dist = pointLineDistance(position(),
			body->position(), body->vectorZ());

		if (dist - ((const GInfEllCylBody*)body)->R() >  R())
			return Location::OUTSIDE;
		else
			return Location::OVERLAP;
	}

	return GBody::_locationWrt(body);
} // _locationWrt

// ================================= GELLBody ==================================
/** set */
void GELLBody::set(const double x1, const double y1, const double z1,
		   const double x2, const double y2, const double z2,
		   const double major)
{
	Point F1(x1, y1, z1);				// focus 1
	Point F2(x2, y2, z2);				// focus 2
	zlen = major/2.0;				// major axis
	if (zlen<=TOOSMALL) throw std::string("Zero or negative major axis");
	P  = 0.5*(F1+F2);				// center
	Z  = F2 - F1;
	double f = Z.normalize()/2.0;			// foci half-distance
	if (major<=f) throw std::string("Full length of major axis smaller than foci distance");
	double minor2 = (zlen-f)*(zlen+f);
	xlen = ylen = (minor2>0.0?Sqrt(minor2):0.0);	// minor axes
	findXYZ();
} // set

/** set */
void GELLBody::set(double* what)
{
	resetBBox();
	round0(what, nwhats(), SMALL);
	set(what[0], what[1], what[2], what[3], what[4], what[5], what[6]);
} // set

/* --- get --- */
int GELLBody::get(double* what) const
{
	resetBBox();
	double f = Sqrt((zlen+xlen)*(zlen-xlen));
	Point F1 = P - f*Z;
	Point F2 = P + f*Z;
	what[0] = ROUND2INT(F1.x, SMALL1);
	what[1] = ROUND2INT(F1.y, SMALL1);
	what[2] = ROUND2INT(F1.z, SMALL1);
	what[3] = ROUND2INT(F2.x, SMALL1);
	what[4] = ROUND2INT(F2.y, SMALL1);
	what[5] = ROUND2INT(F2.z, SMALL1);
	what[6] = 2.0*zlen;
	return 7;
} // get

/** @return transformed position */
Point GELLBody::node(const int n) const
{
	double f = Sqrt((zlen+xlen)*(zlen-xlen));
	Point r;
	switch (n) {
		case 0: r = P;
			break;
		case 1: r = P - f*Z;
			break;
		case 2: r = P + f*Z;
			break;
	}
	return _hasMatrix? _matrix*r : r;
} // node

/**  @return closest point, edge, face */
int GELLBody::closest(const Point& r, const double d, const Vector& w) const
{
	Point tr = _hasMatrix? _invMatrix*r : r;

	double f = Sqrt((zlen+xlen)*(zlen-xlen));
	if ((P-f*Z - tr).length2() < Sq(d)) return 16;
	if ((P+f*Z - tr).length2() < Sq(d)) return 17;

	int close = GBody::closest(r,d,w);
	if (close == 1) {
		double z2 = Sq(tr*SZ);
		double xy2 = tr.length2() - z2;
		if (xy2>z2) close = 2;
	}
	return close;
} // closest

/** move body item to location
 * @param item	item to move as returned by closest
 * @param r	position vector
 * @param w	vertical axis of the viewport
 */
void GELLBody::move(int item, const Point& r, const Vector& w)
{
	resetBBox();
	Point tr = _hasMatrix? _invMatrix*r : r;
	if (item==1) {
		tr -= SP;
		double z2 = Sq(tr*SZ);
		double xy2 = tr.length2() - z2;
		double a = z2 / (1.0-xy2/Sq(sxlen));
		zlen = a>0.0? Sqrt(a) : szlen;
		// Check if we have a valid ELL
		if (zlen < xlen) zlen = szlen;
	} else
	if (item==2) {
		tr -= SP;
		double z2 = Sq(tr*SZ);
		double xy2 = tr.length2() - z2;
		double a = xy2 / (1.0-z2/Sq(szlen));
		xlen = ylen = a>0.0? Sqrt(a) : sxlen;
		if (zlen < xlen) xlen = ylen = sxlen;
	} else
	if (item==16) {
		double what[7];
		get(what);
		what[0] = tr.x;
		what[1] = tr.y;
		what[2] = tr.z;
		set(what);
	} else
	if (item==17) {
		double what[7];
		get(what);
		what[3] = tr.x;
		what[4] = tr.y;
		what[5] = tr.z;
		set(what);
	} else
		GBody::move(item, r, w);
} // move

/** createQuads */
void GELLBody::createQuads()
{
	GBody::createQuads();
	if (ISZERO(xlen) || ISZERO(ylen) || ISZERO(zlen)) return;

	// add quad centered to origin
	//      Cxx        Cyy     Czz     Cxy  Cxz  Cyz  Cx   Cy   Cz     C
	addQuad(1.0/Sq(xlen), 1.0/Sq(ylen), 1.0/Sq(zlen), 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -1.0 );

	// rotate to align to F1-F2 direction
	Matrix4 T, M, Minv;
	Minv.make(X,Y,Z);		// rotation matrix
	Minv.inverse();
	T.translate(P);			// translation to P
	M.multiply(T, Minv);
	Minv.inverse(M);

	// transform first quad
	_Q[0].transform(Minv, M);
	_Q[0].normalize();
} // createQuads

/** @return bounding box of body */
void GELLBody::updateBBox() const
{
	_bbox = Q(0).bbox();
} // bbox

#ifdef OBBOX
/** @return oriented bounding box of body */
void GELLBody::updateOBB(bool in) const
{
	OBBox& bb = _obbox[(int)in];

	bb.P = GBody::position();
	bb.X = vectorX();
	bb.Y = vectorY();
	bb.Z = vectorZ();

	if (!in) {
		bb.low()  = Point(-xlen, -ylen, -zlen);
		bb.high() = Point( xlen,  ylen,  zlen);
	} else {
		double sc = 1.0/Sqrt3;
		bb.low()  = Point(-xlen*sc, -ylen*sc, -zlen*sc);
		bb.high() = Point( xlen*sc,  ylen*sc,  zlen*sc);
	}
} // updateOBB
#endif
