/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	09-Jan-2008
 */

#ifndef __MATH_BBOX_H
#define __MATH_BBOX_H

#include <ostream>

#include "eps.h"
#include "point.h"

class	Quad;
class	Matrix4;

/** BBox - bounding box class */
class BBox {
public:
	/** caching bbox possibilities */
	enum class State {
		Invalid,	/* needs re-calculation	*/
		Valid,		/* bbox is ok		*/
		UserDefined	/* user defined bbox	*/
	};

private:
	Point	lowPt;
	Point	highPt;
	bool	valid;

public:
	/** Empty bounding box */
	BBox(bool inf=false)	{
			if (inf)
				infinite();	// create an infinite bounding box
			else
				reset();	// Create a Non valid bounding box
		}

	BBox(const double xmin, const double xmax,
	     const double ymin, const double ymax,
	     const double zmin, const double zmax)
		: lowPt(xmin,ymin,zmin),
		  highPt(xmax,ymax,zmax)
		{ checkValid(); }

	/** Create a bounding box from low&high point
	 * @param aLow low point
	 * @param aHigh low point
	 */
	BBox(const Point& aLow, const Point& aHigh)
		: lowPt(aLow),
		  highPt(aHigh)
		{ checkValid(); }

	/** Copy constructor */
	BBox(const BBox&) = default;

	// Assignment
	BBox&	operator =(const BBox&) = default;

	/** reset bounding box to invalid */
	void	reset() {
			lowPt.set(  INFINITE,  INFINITE,  INFINITE);
			highPt.set(-INFINITE, -INFINITE, -INFINITE);
			valid = false;
		}

	/** set bounding box to infinite */
	void	infinite() {
			lowPt.set(-INFINITE, -INFINITE, -INFINITE);
			highPt.set(INFINITE,  INFINITE,  INFINITE);
			valid = true;
		}

	/** set bounding box to null */
	void	null() {
			lowPt.set(0, 0, 0);
			highPt.set(0, 0, 0);
			valid = true;
		}

	/** check if bounding box is valid
	 * @return true if lowPt < highPt
	 */
	bool	isValid() const { return valid; }
	bool	isInfinite() const {
			return	lowPt.x==-INFINITE and
				lowPt.y==-INFINITE and
				lowPt.z==-INFINITE and
				highPt.x==INFINITE and
				highPt.y==INFINITE and
				highPt.z==INFINITE;
		}

	/** @return low point */
	Point	low()	const	{ return lowPt; }
	/** @return high point */
	Point	high()	const	{ return highPt; }

	/** set low point
	 * @param point to be set as low
	 */
	void	low(const Point& point)		{ lowPt = point; checkValid(); }
	void	low(const double x, const double y, const double z) {
				lowPt.x = x;
				lowPt.y = y;
				lowPt.z = z;
				checkValid();
			}
	void	lowx(const double x)	{
				lowPt.x = x;
				checkValid();
			}
	void	lowy(const double y)	{
				lowPt.y = y;
				checkValid();
			}
	void	lowz(const double z)	{
				lowPt.z = z;
				checkValid();
			}
	double	lowx() const	{ return lowPt.x; }
	double	lowy() const	{ return lowPt.y; }
	double	lowz() const	{ return lowPt.z; }

	/** set high point
	 * @param point to be set as high
	 */
	void	high(const Point& point)	{ highPt = point; checkValid(); }
	void	high(const double x, const double y, const double z) {
				highPt.x = x;
				highPt.y = y;
				highPt.z = z;
				checkValid();
			}
	void	highx(const double x)	{
				highPt.x = x;
				checkValid();
			}
	void	highy(const double y)	{
				highPt.y = y;
				checkValid();
			}
	void	highz(const double z)	{
				highPt.z = z;
				checkValid();
			}
	double	highx() const	{ return highPt.x; }
	double	highy() const	{ return highPt.y; }
	double	highz() const	{ return highPt.z; }

	/** @return size of bbox */
	Vector	size() const	{ return highPt - lowPt; }

	/** @return volume of bounding box */
	double	volume() const	{ Vector d=size(); return d.x*d.y*d.z; }

	/** @return surface of bounding box, needed for SAH bbox creating */
	double  surface() const {
			Vector d = size();
			return 2.0 * (d.x*d.y + d.y*d.z + d.z*d.x);
		}

	/** @return centroid point */
	Point  centroid() const {
			return lowPt + (highPt - lowPt) / 2.0;
		}

	/** @return projection of centroid on a specific axis */
	double centroid(int axis) const {
			assert(axis>=0 && axis<3 );
			return lowPt[axis] + (highPt[axis] - lowPt[axis])/2.0;
		}

	/** Add point
	 * @param x,y,z to be added
	 */
	void	add(const double x, const double y, const double z);

	/** Add point
	 * @param p	point to be added
	 */
	void	add(const Point& p)	{ add(p.x, p.y, p.z); }

	/** vertices
	 * @return number of vertices
	 */
	int	vertices()	const { return 8; }

	/** vertex(i)
	 * @return one of the numbered vertices
	 */
	Point	vertex(const int i) const;

	/** edges
	 * @return number of edges
	 */
	int	edges()		const { return 12; }

	/** edge(i)
	 * @return one of the numbered edges
	 */
	void	edge(const int i, int* a, int* b) const;

	/** Transform the bounding box
	 * @param trans transformation to be used
	 */
	void	transform(const Matrix4& matrix);

	/** Test for overlapping 2 bounding boxes
	 * @param b second bounding box
	 * @param eps accuracy for testing
	 * @return true if the overlap within accuracy eps
	 */
	bool	overlap(const BBox& b, double eps=TOOSMALL) const;

	/** Test is a specified point lies inside a bounding box
	 * @param point to test
	 * @param eps accuracy for testing
	 * @return true is point is inside within accuracy eps
	 */
	bool	inside(const double x, const double y,
			const double z, const double eps=TOOSMALL) const;
	bool	inside(const Point& point, const double eps=TOOSMALL) const
			{ return inside(point.x, point.y, point.z, eps); }

	/** Create the union with another bounding box */
	void	Union(const BBox& b);

	/** Create the intersection with another bounding box */
	void	Intersect(const BBox& b);

	/** Create the difference with another bounding box
	 * WARNING: this works only if the bbox is equal to the body!
	 *          e.g untransformed RPP, XYP, YZP, XZP
	 */
	void	differenceEqual(const BBox& b);

	/** Intersect with a plane (positive or negative)
	 * @return true if intersected
	 */
	bool	intersectPlane(const Quad& plane, bool negative);

	/** Check for intersection distance against a ray
	 * @param x,y,z		initial position (I)
	 * @param dx,dy,dz	direction (I)
	 * @param tmin,tmax	minimum( and maximum search distance
	 * @return true is ray intersects the bounding box
	 */
	bool	distance(const double  x, const double  y, const double  z,
			 const double dx, const double dy, const double dz,
			 const double tmin, const double tmax) const;

	bool	distance(const Point& p, const Vector& d,
			 const double tmin, const double tmax) const
				{ return distance(p.x,p.y,p.z, d.x,d.y,d.z, tmin, tmax); }

	// bounding box operations
	BBox& operator |= (const BBox& b) {
			Union(b);
			return *this;
		}

	BBox& operator += (const BBox& b) {
			Intersect(b);
			return *this;
		}

private:
	void	checkValid() {
			valid = (lowPt.x <= highPt.x) and
				(lowPt.y <= highPt.y) and
				(lowPt.z <= highPt.z);
		}
}; // BBox

/** transform a bounding box */
BBox operator * (const Matrix4&, const BBox&);

/** print on a stream a bounding box */
std::ostream& operator << (std::ostream& out, const BBox& obj);

#endif
