/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	02-Feb-2015
 */

#ifndef __MATH_HISTOGRAM_H
#define __MATH_HISTOGRAM_H

#include <assert.h>
#include <string>

// ================================= Histogram =================================
class Histogram {
private:
	std::string _title;	/** histogram title		*/

	int	xbins;		/** number of bins		*/
	double	xlow;		/** lower x limit		*/
	double	xhigh;		/** higher x limit		*/
	double	xstep;		/** x step size			*/

	int	_entries;	/** number of entries		*/
	double	_under;		/** weights below low x limit	*/
	double	_over;		/** weights above high x limit	*/
//	double	_total;		/** total sum of weights	*/

	double	_sumw;		/** sum of weights		*/
	double  _sumw2;		/** sum of weights^2		*/
	double	_sumwx;		/** sum of weight*x		*/
	double	_sumwx2;	/** sum of weight*x**2		*/

	double	*h;		/** histogram values		*/
	double	*eh;		/** errors			*/

public:
	Histogram(const char *atitle=NULL, int bins=0, double low=0.0, double high=1.0);
	Histogram(const Histogram& hist);
	Histogram& operator =(const Histogram&) = default;
	~Histogram();

	/* initialize */
	void	free();
	void	clear();

	/* get/set */

	void	title(const std::string& t)	{ _title = t; }
	void	title(const char *t)	{ _title = t; }
const	std::string&	title()	const	{ return _title; }
const	char*	ctitle()	const	{ return _title.c_str(); }

	void	set(int n, double low, double high);
	void	setLimits(double low, double high);

	int	nbins()		const	{ return xbins;    }
//	int	nbinsx()	const	{ return xbins;    }
	int	entries()	const	{ return _entries; }
	double	under()		const	{ return _under;   }
	double	over()		const	{ return _over;    }
	double	sumw()		const	{ return _sumw;    }
	double	sumw2()		const	{ return _sumw2;   }
	double	sumwx()		const	{ return _sumwx;   }
	double	sumwx2()	const	{ return _sumwx2;  }
	double	width()		const	{ return xstep;    }
	double	mean()		const	{ return _sumwx / _sumw; }

	/* fill histogram */
	int	geti(const double x)	const	{ return (int)((x - xlow) / xstep); }
	double	getx(const int i)	const	{ return (double)i*xstep + xlow; }
	double	center(const int i)	const	{ return xlow + i*xstep + 0.5*xstep; }
	int	fill(const double x, const double w=1.0);

	void	set(int bin, double content);
	double	get(int i) const	{ assert(i>=0 && i<xbins); return h[i]; }
	double	operator[](int i) const	{ assert(i>=0 && i<xbins); return h[i]; }
	double&	operator[](int i)	{ assert(i>=0 && i<xbins); return h[i]; }

	double	error(int i)	const	{ assert(i>=0 && i<xbins); return eh[i]; }

	void	norm(const double f);
	void	scale(const double f)	{ norm(f); }
	void	isolethargic();
	void	cumulative(bool reverse=false);

	/* I/O */
	void	save(FILE *fout, int columns=4, bool lastdup=false);
	bool	load(FILE *fin);
}; // Histogram

// ==================================== H1D ====================================
class H1D : public Histogram {
public:
	H1D(const char *atitle=NULL, int bins=0, double low=0.0, double high=1.0)
		: Histogram(atitle, bins, low, high) {}
	H1D(const H1D& hist) : Histogram(hist) {}
	H1D& operator =(const H1D&) = default;
}; // class H1D

#endif
