/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 */

#include "usrbinlayer.h"
#include "viewer.h"

UsrbinLayer::UsrbinLayer(const Geometry& g, GeometryKernel& k, GeometryViewer& v) : Layer(g, k, v)
{
	for (int i=0; i<MAXUSRBIN; i++) {
		_alpha[i]   = 0;
		_palette[i] = 0;
	}
} // UsrbinLayer

/** draw
 * FIXME 1. Can be optimized by introducing a bounding box around the image and calculate
 *          first the intersection of the 6 planes
 *       2. Modified to check for non-lines, in order to draw also the regions
 */
void UsrbinLayer::draw(Painter& painter)
{
	Point	r(0.0, 0.0, 0.0);
	int H = painter.height();	// save variables that might change
	int W = painter.width();	// during the scanning...
	Color32* ptr = (Color32*)painter.data();	// data pointer

	int maxusrbin = -1;
	for (int usrbinidx=0; usrbinidx < MAXUSRBIN; usrbinidx++) {
		Usrbin &usrbin = usrbins[usrbinidx];
		if (!usrbin.hasData() && !usrbin.checker()) continue;
		maxusrbin = usrbinidx;
	}
	if (maxusrbin<0) return;

	for (int j=0; j<H; j++) {
		if (stop()) break;
		r.y = view().jc2v(j);

		for (int i=0; i<W; i++, ptr++) {
			dword flag = *ptr & FLAG_INFOMASK;
			if (flag==0) continue;
			if ((viewer.d3.show && flag!=FLAG_REGION) ||
			    flag&(FLAG_FINAL|FLAG_ALPHA)) continue;
			r.x = view().ic2u(i);

			for (int usrbinidx=0; usrbinidx <= maxusrbin; usrbinidx++) {
				Usrbin& usrbin = usrbins[usrbinidx];
				if (!usrbin.hasData() && !usrbin.checker()) continue;
				Point p = view().matrix() * r;

				double value;
				if (!usrbin.getData(p.x, p.y, p.z, &value)) continue;

				Color32 usr;
				int pal = _palette[usrbinidx];
				if (usrbin.checker()) {
					if (value<0.5)
						usr = viewer.palette[pal].first();
					else
						usr = viewer.palette[pal].last();
				} else {
					dword color = viewer.palette[pal].color(value);
					if (color == COLOR_TRANSPARENT)
						break;
					else
						usr = color;
				}
				ptr->alpha() = alpha(usrbinidx);
				usr.alpha()  = 0xFF;
				ptr->blend(usr);
			}
		}
	}
} // draw

/** shade
 * @param hit	input 3D hit position
 * @param color output color at hit position
 * @return 0xFF if no usrbin is found at hit position otherwise 0..255 alpha level
 */
byte UsrbinLayer::shade(const Point& hit, Color& color) const
{
	// move a bit the hit position to avoid boundary effects
	for (int i=0; i<MAXUSRBIN; i++) {
		const Usrbin& bin = usrbins[i];
		if (!bin.hasData() && !bin.checker()) continue;

		double value;
		if (!bin.getData(hit.x, hit.y, hit.z, &value)) continue;

		Color usrcolor;
//			color->rgb.alpha = alpha(i);
		int pal = _palette[i];
		dword c;
		if (bin.checker()) {
			if (value<0.5)
				c = viewer.palette[pal].first();
			else
				c = viewer.palette[pal].last();
		} else {
			c = viewer.palette[pal].color(value);
			if (c == COLOR_TRANSPARENT) break;
		}
		usrcolor.set(c);
//			usr.alpha = 0xFF;
		color.blend(usrcolor, (float)alpha(i)/255.0f);
//			color->val = alphaBlend(*color, usr);
		return alpha(i);
	}
	return 0xFF;
} // shade

/** regionColorFromUsrbin
 * Set color of all regions using the USRBIN values
 * if not defined set to 0
 */
void UsrbinLayer::regionColorFromUsrbin(int idx)
{
	if (!usrbins[idx].hasData()) return;
	int pal = _palette[idx];
	for (auto region : kernel.regions) {
		region->color(viewer.palette[pal].first());
		double value;
		if (usrbins[idx].getData(region->id(),&value))
			region->color(viewer.palette[pal].color(value));
	}
} // regionColorFromUsrbin
