/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 */

#ifndef __VIEWER_GEOMETRY_VIEWER_H
#define __VIEWER_GEOMETRY_VIEWER_H

#include "d2layer.h"
#include "d3layer.h"
#include "imagelayer.h"
#include "voxellayer.h"
#include "exportlayer.h"
#include "usrbinlayer.h"
#include "latticelayer.h"
#include "palettelayer.h"
#include "userdumplayer.h"
#include "decorationlayer.h"

#include "math/bit.h"
#include "graph/zpainter.h"

class GeometryKernel;

// Selection mask bits (synchronize with flair)
enum SelectionMask {
	//BIT_INVALID     = (dword)-1,
	BIT_SELECT	= BIT0,
	BIT_FREEZE	= BIT1,
	BIT_LOCK	= BIT2,
	BIT_VISIBLE	= BIT3,
	BIT_WIREFRAME	= BIT4,
	BIT_MOVE	= BIT5,
	BIT_BBOX	= BIT6,

	BIT_SELFREEZE	= BIT_SELECT | BIT_FREEZE | BIT_VISIBLE,
	BIT_SELBBOX	= BIT_SELECT | BIT_BBOX,
};

// Drawing mask bits
enum DrawMask {
	DRAW_CLEAR	= 1,
	DRAW_SEGMENTS	= 1<<1,
	DRAW_REGIONS	= 1<<2,
	DRAW_GRID	= 1<<3,
	DRAW_AXES	= 1<<4,
	DRAW_TITLE	= 1<<5,
	DRAW_STATUS	= 1<<6,
	DRAW_LABELS	= 1<<7,
	DRAW_IMAGE	= 1<<8,
	DRAW_USRBIN	= 1<<9,
	DRAW_3D		= 1<<10,
	DRAW_EDGE	= 1<<11,
	DRAW_VERTICES	= 1<<12,
	DRAW_USERDUMP	= 1<<13,
	DRAW_PALETTE	= 1<<14,
	DRAW_LATTICES	= 1<<15,
	DRAW_VOXEL	= 1<<16,
	DRAW_WIREFRAME	= 1<<17,
	DRAW_BBOX	= 1<<18,
	DRAW_MESSAGE	= 1<<19,
//	DRAW_TEST	= 1<<30
};

typedef void (*NotifyFunc)(void *arg);

/* ========================== GeometryViewer ========================== */
/** Geometry Viewer class */
class GeometryViewer {
public:
	// Describe current step of the Geometry::project() method
	// XXX should be in sync with GeometryViewer.py
	enum class Projection : int {
		NOTHING   = 0,	// No projection information
		SPAWN     = 1,	// spawn projection (waiting to start)
		START     = 2,	// start projection
		CONICS    = 3,	// Transform bodies and prepare conics
		LOCATION  = 4,	// Calculate the bodies&region location
		INTERSECT = 5,	// Intersect all bodies
		SCAN      = 6,	// Scan segments
		DRAW      = 7,	// Late drawing as a thread
		FINISHED  = 10	// Projection ended
	};

protected:
const   Geometry&	geometry;	/** father geometry object	*/
	GeometryKernel&	kernel;		/** geometry kernel with engines*/
	std::string	_title;		/** view title			*/

public:
	bool		showErrors;	/** flag ON/OFF for errors	*/
	int	textBackgroundLevel;	/** text background intensity	*/
	ZPainter	painter;	/** painter class		*/

public:					// Layers
	DecorationLayer decoration;	/** decoration layer		*/
	PaletteLayer	palette;	/** Color box layer		*/
	D2Layer		d2;		/** 2D layer			*/
	D3Layer		d3;		/** 3D layer			*/
	LatticeLayer	lattice;	/** lattice layer		*/
	VoxelLayer	voxel;		/** voxel layer			*/
	UserDumpLayer	userdump;	/** userdump layer		*/
	UsrbinLayer	usrbin;		/** usrbin layer		*/
	ImageLayer	image;		/** image layer			*/
	ExportLayer	exporter;	/** export layer		*/
	BFont		font;		/** general font		*/

protected:				// Layers
	// Treading variables
	pthread_t	thread;		/** thread handler		*/
	pthread_mutex_t	mutex;		/** main viewer mutex		*/
	pthread_mutex_t	mutexStop;	/** stop thread mutex		*/
	pthread_mutex_t	mutexDraw;	/** mutex for late drawing	*/
volatile bool		pStop;		/** stop signal			*/
volatile Projection	pState;		/** projection state		*/
	Projection	_saveState;	/** saved state for draw	*/
	int		_percent;	/** projection percentage	*/
	VBody*		pBody;		/** last processed body		*/
	NotifyFunc	endThread;	/** end project notification	*/
	void*		endArg;		/** arguments to notify	func	*/
	int		_mask;		/** spawned arguments		*/
	int		_fast;		/** spawned arguments		*/

public:
	GeometryViewer(const Geometry& g, GeometryKernel& k);
	~GeometryViewer();

	void	makeLatticeColor();

	/* set/get */
	void	title(const std::string& t){ _title = t; }
const	std::string&	title()	const	{ return _title; }

	void	percent(int p)		{ _percent = p; }
	int	percent()	const	{ return _percent; }

	/** window */
const	ViewPort& view()	const	{ return kernel.view; }
	ViewPort& view()		{ return kernel.view; }
	void	window(double ex, double ey=0.0);
	void	window(double xmin, double ymin, double xmax, double ymax)
			{ view().window(xmin, ymin, xmax, ymax); }
	void	resize(int w, int h);
	int	width()		const	{ return view().width(); }
	int	height()	const	{ return view().height(); }

	void	backgroundColor(const dword color);

	/* transformations */
	void	matrix(const Matrix3& m);
	void	matrix(const Matrix4& m);
	void	origin(const double x, const double y, const double z);
	void	origin(double *x, double *y, double *z)	{ view().origin(x,y,z); }
	void	moveViewOriginTo0();

	/* threads */
	int	spawnProject(NotifyFunc func=nullptr, void *arg=nullptr);
	int	spawnDraw(NotifyFunc func=nullptr, void *arg=nullptr, int mask=-1,
				bool fast=false);
	void	stopThread();

	Projection	state()	const	{ return pState; }
	bool	stop()		const	{ return pStop; }

	/* drawing */
	Projection	draw(int mask=-1, bool fast=false);

	/* Info */
	size_t	memory()	const;
	void	printMemory()	const;

	void	lock()			{ pthread_mutex_lock(&mutex); }
	void	unlock()		{ pthread_mutex_unlock(&mutex); }

protected:
	void	lockStop()		{ pthread_mutex_lock(&mutexStop); }
	void	unlockStop()		{ pthread_mutex_unlock(&mutexStop); }
	void	lockDraw()		{ pthread_mutex_lock(&mutexDraw); }
	void	unlockDraw()		{ pthread_mutex_unlock(&mutexDraw); }

	bool	state(Projection s, int p);
	void	saveState()		{ _saveState = pState; }
	void	restoreState()		{ pState = _saveState; }
static	void*	runProject(GeometryViewer *gthis);
static	void*	runDraw(GeometryViewer *gthis);

friend class Layer;
friend class D2Layer;
}; // GeometryViewer
#endif
